use git2::Repository;
use std::fs::{self, File};
use std::io::{self, Read, Write};
use std::path::Path;
use toml::Value;
use yaml_rust::YamlLoader;

pub fn download_sources(data: &Path, store: &Path, cache: &Path) -> Result<(), io::Error> {

    if !data.exists() {

        fs::create_dir_all(data.parent().unwrap())?; // The unwraps should not be panicking

        let mut file = File::create(data)?;

        file.write_all(
            b"[schemes]
local = [ \"~/.config/themectl/schemes\" ]
sources = [ \"https://github.com/chriskempson/base16-schemes-source\" ]

[templates]
local = [ \"~/.config/themectl/templates\" ]
sources = [ \"https://github.com/chriskempson/base16-templates-source\" ]
",
        )?;
    }

    if !store.exists() {
        fs::create_dir_all(store)?;
        fs::create_dir_all(store.join("schemes"))?;
        fs::create_dir_all(store.join("templates"))?
    }


    let data = fs::read_to_string(data)?
        .parse::<Value>()?;

    for source in data
        .get("schemes")
        .unwrap_or(&Value::String("".to_string()))
        .get("sources")
        .unwrap_or(&Value::Array([].to_vec()))
        .as_array()
        .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
        .iter()
    {

        let src = source
            .as_str()
            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?;

        println!("Retrieving from {}...", src);

        git_get(src, &cache.join("schemes"))?;

        let yml = yaml_from_file(&cache.join("schemes/list.yaml"))?;

        for (_name, url) in yml[0]
            .as_hash()
            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect source list format"))?
            .iter() {

            git_get(url.as_str().ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect url for source."))?,
                &cache.join("gitschemes"))?;

            let dict = fs::read_dir(&cache.join("gitschemes"))?
                    .into_iter()
                    .filter(|x|
                            ["yml", "yaml"].contains(&(match x.as_ref().unwrap().path().extension() {
                                Some(a) => a.to_str().unwrap(),
                                None => ""                                                  // All these unwraps shouldn't panic, if they do something is very wrong
                            })))
                    .map(|x| x.unwrap().path().into_os_string().into_string().unwrap())
                    .collect::<Vec<String>>();

            for path in dict {
                if fs::copy(
                    path.clone(),
                    store
                        .join("schemes")
                        .join(Path::new(&path).file_name().unwrap()),
                )
                .is_ok()
                {};
            }
        }
    }

    for source in data
        .get("templates")
        .unwrap_or(&Value::String("".to_string()))
        .get("sources")
        .unwrap_or(&Value::Array([].to_vec()))
        .as_array()
        .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
        .iter()
    {

        let src = source
            .as_str()
            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?;

        println!("Retrieving from {}...", src);

        git_get(src, &cache.join("template"))?;

        let yml = yaml_from_file(&cache.join("template/list.yaml"))?;

        for (name, url) in yml[0]
            .as_hash()
            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect source list format"))?
            .iter() {


            git_get(url.as_str().ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect url for source."))?,
                &cache.join("gittemplates"))?;

            let dict = fs::read_dir(&cache.join("gittemplates/templates"))?
                    .into_iter()
                    .filter(|x|
                            ["mustache"].contains(&(match x.as_ref().unwrap().path().extension() {
                                Some(a) => a.to_str().unwrap(),
                                None => ""                                                  // All these unwraps shouldn't panic, if they do something is very wrong
                            })))
                    .map(|x| x.unwrap().path().into_os_string().into_string().unwrap())
                    .collect::<Vec<String>>();

            for path in dict {
                if fs::copy(
                    path.clone(),
                    store
                        .join("templates")
                        .join(format!("{}-{}", name
                                      .as_str()
                                      .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect template format"))?, 
                            Path::new(&path).file_name().unwrap().to_str().unwrap()
                            ).replace("-default", "")))
                .is_ok()
                {};
                
                
            }

        }
    }

    Ok(())
}

pub fn clean(store: &Path) -> Result<(), io::Error> {
    fs::remove_dir_all(store.join("schemes"))?;
    fs::remove_dir_all(store.join("templates"))?;
    fs::create_dir_all(store.join("schemes"))?;
    fs::create_dir_all(store.join("templates"))?;

    Ok(())
}

fn git_get(url: &str, path: &Path) -> Result<(), io::Error> {
    let dir = path.as_os_str().to_str().ok_or_else(|| io::Error::new(io::ErrorKind::NotFound, "Path not found."))?;

    if fs::remove_dir_all(dir).is_ok() {};
    Repository::clone(url, dir).expect("Failed to clone. ");

    Ok(())
}

pub fn yaml_from_file(file: &Path) -> Result<Vec<yaml_rust::Yaml>, io::Error> {
    let mut f = File::open(file)?;
    let mut string = String::new();
    f.read_to_string(&mut string).unwrap();
    match YamlLoader::load_from_str(&string) {
        Ok(x) => Ok(x),
        Err(_) => Err(io::Error::new(io::ErrorKind::InvalidData, "Invalid yaml."))
    }
}

