use std::collections::HashMap;
use std::fs;
use std::fs::File;
use std::io::BufWriter;
use std::path::Path;
use std::process::Command;
use toml::{Value, map::Map};
use rustache::{HashBuilder, Render};
use crate::builder::{yaml_from_file};
use std::io;

pub fn apply(config: &Path, store: &Path, data: &Path, source: &Option<String>) -> Result<(), io::Error> {
    match source {
        Some(st) => {
            apply_theme(config, store, data, st.to_string())?;
        }

        None => {
            list(data, store, &None, "schemes")?;
        }
    }

    Ok(())
}

pub fn list(data: &Path, store: &Path, source: &Option<String>, typ: &str) -> Result<(), io::Error> {

    let data = fs::read_to_string(data)?
        .parse::<Value>()?;

    match source {

        Some(st) => {
            if st == "sources" {

                println!("Source {}: ", typ);

                let dict = fs::read_dir(&store.join(typ))?
                    .map(|x| String::from(x.unwrap().path().file_stem().unwrap().to_str().unwrap())) // This shouldn't error
                    .collect::<Vec<String>>();

                for i in dict {
                    println!(" - {}", i);
                }

            } else if st == "local" {

                println!("Custom {}: ", typ);

                for local in data
                    .get(typ)
                    .unwrap_or(&Value::Table(Map::new()))
                    .get("local")
                    .unwrap_or(&Value::Array([].to_vec()))
                    .as_array()
                    .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
                    .iter()
                {

                    let dict = fs::read_dir(Path::new(
                        &local.as_str().unwrap().replace("~", env!("HOME")),
                    ))?
                    .map(|x| String::from(x.unwrap().path().file_stem().unwrap().to_str().unwrap()))
                    .collect::<Vec<String>>();

                    for i in dict {
                        println!(" - {}", i);
                    }
                }

            } else {
                println!("Please use a valid option (sources, local). ");
            }
        }
        
        None => {

            println!("Source {}: ", typ);

            let dict = fs::read_dir(&store.join(typ))?
                .map(|x| String::from(x.unwrap().path().file_stem().unwrap().to_str().unwrap()))
                .collect::<Vec<String>>();

            for i in dict {
                println!(" - {}", i);
            }

            println!("\nCustom {}: ", typ);

            for local in data
                .get(typ)
                .unwrap_or(&Value::Table(Map::new()))
                .get("local")
                .unwrap_or(&Value::Array([].to_vec()))
                .as_array()
                .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
                .iter()
            {

                let dict = fs::read_dir(Path::new(
                    &local.as_str().unwrap().replace("~", env!("HOME")),
                ))?
                .map(|x| String::from(x.unwrap().path().file_stem().unwrap().to_str().unwrap()))
                .collect::<Vec<String>>();

                for i in dict {
                    println!(" - {}", i);
                }

            }
        }
    }
    Ok(())
}

fn apply_theme(config: &Path, store: &Path, data: &Path, theme: String) -> Result<(), io::Error> {

    let config = fs::read_to_string(config)?
        .parse::<Value>()?;

    let data = fs::read_to_string(data)?
        .parse::<Value>()?;

    let mut templates: HashMap<String, String> = HashMap::new();

    for local in data
        .get("templates")
        .unwrap_or(&Value::Table(Map::new()))
        .get("local")
        .unwrap_or(&Value::Array([].to_vec()))
        .as_array()
        .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
        .iter()
    {
        let dict = fs::read_dir(Path::new(
            &local.as_str().unwrap().replace("~", env!("HOME")),
        ))?
        .map(|x| String::from(x.unwrap().path().to_str().unwrap()))
        .collect::<Vec<String>>();

        for i in dict {
            templates.insert(Path::new(&i).file_stem().unwrap().to_str().unwrap().to_string(),
                             i.clone());
        }

    }

    let dict = fs::read_dir(&store.join("templates"))?
        .map(|x| String::from(x.unwrap().path().to_str().unwrap())) // This shouldn't error
        .collect::<Vec<String>>();


    for i in dict {
        templates.insert(Path::new(&i).file_stem().unwrap().to_str().unwrap().to_string(),
                         i.clone());
    }

    let mut schemes: HashMap<String, String> = HashMap::new();

    for local in data
        .get("schemes")
        .unwrap_or(&Value::Table(Map::new()))
        .get("local")
        .unwrap_or(&Value::Array([].to_vec()))
        .as_array()
        .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Incorrect sources format"))?
        .iter()
    {
        let dict = fs::read_dir(Path::new(
            &local.as_str().unwrap().replace("~", env!("HOME")),
        ))?
        .map(|x| String::from(x.unwrap().path().to_str().unwrap()))
        .collect::<Vec<String>>();

        for i in dict {
            schemes.insert(Path::new(&i).file_stem().unwrap().to_str().unwrap().to_string(),
                             i.clone());
        }

    }

    let dict = fs::read_dir(&store.join("schemes"))?
        .map(|x| String::from(x.unwrap().path().to_str().unwrap())) // This shouldn't error
        .collect::<Vec<String>>();


    for i in dict {
        schemes.insert(Path::new(&i).file_stem().unwrap().to_str().unwrap().to_string(),
                         i.clone());
    }

    let slug: String = schemes.get(&theme)
        .ok_or_else(|| io::Error::new(io::ErrorKind::NotFound, "Theme not found."))?
        .clone();

    fs::copy(Path::new(&slug), &store.join("current.yaml"))?;

    let yaml = &yaml_from_file(Path::new(&slug))?[0];
    let mut name = String::new();
    let mut author = String::new();

    let mut colors: HashMap<String, String> = HashMap::new();

    for (a, v) in yaml
        .as_hash()
        .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Error reading theme file."))?
        .iter() {

            let val = v.as_str()
                .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, "Error reading theme file."))?
                .to_string();

            match a.as_str().unwrap() {
                "scheme" => name = val,
                "author" => author = val,

                _ => {
                    colors.insert(a.as_str().unwrap().to_string(), val);
                }
            }
    }


    if Command::new("sh")
        .arg("-c")
        .arg(config.get("pre")
        .unwrap_or(&Value::String("".to_string()))
        .as_str().unwrap_or(""))
        .spawn()
        .is_ok()
    {};

    for (key, options) in config.get("apps")

        .unwrap_or(&Value::Table(Map::new()))
        .as_table()
        .unwrap()
        .into_iter() {

            let key = key.as_str();
            let file = match options.as_table() {
                Some(v) => v.get("file")
                    .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, format!("Couldn't find file option for {}",key)))?
                    .as_str()
                    .unwrap()
                    .replace("~", env!("HOME")),
                None => options.as_str()
                    .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, format!("Couldn't find file option for {}",key)))?
                    .replace("~", env!("HOME")),
            };

        // We have key -> Here check if key exists as a name in templates variable (String: String)
        // where it is TemplateName: TemplateFile

        templates.keys()
            .find(|&x| key == x)
            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidData, format!("No key found: {}", key)))?;

        if Command::new("sh")
            .arg("-c")
            .arg(options
                .as_table()
                .unwrap_or(&Map::new())
                .get("pre")
                .unwrap_or(&Value::String("".to_string()))
                .as_str()
                .unwrap_or(""))
            .spawn()
            .is_ok()
        {};

        // Here we want to parse the mustache file, and write it to the final file
        // (open template file + save mustache to string)
        // Copy in base16 colors in
        // Write to `file` variable
        
        let mustache = fs::read_to_string(templates.get(key).unwrap())?; // shouldn't error

        // Need to detect these errors
        let mut musthash = HashBuilder::new();
        musthash = musthash.insert("scheme-slug", slug.as_ref());
        musthash = musthash.insert("scheme-name", name.as_ref());
        musthash = musthash.insert("scheme-author", author.as_ref());
        for (b, c) in &colors {
            musthash = musthash.insert(b.to_string() + "-hex", c.as_ref());

            let red = c[0..2].to_string();
            musthash = musthash.insert(b.to_string() + "-hex-r", red.as_ref());
            let red = i32::from_str_radix(c[0..2].as_ref(), 16).unwrap();
            musthash = musthash.insert(b.to_string() + "-rgb-r", red);
            musthash = musthash.insert(b.to_string() + "-dec-r", red as f64 / 255.0);

            let green = c[2..4].to_string();
            musthash = musthash.insert(b.to_string() + "-hex-g", green.as_ref());
            let green = i32::from_str_radix(c[2..4].as_ref(), 16).unwrap();
            musthash = musthash.insert(b.to_string() + "-rgb-g", green);
            musthash = musthash.insert(b.to_string() + "-dec-g", green as f64 / 255.0);

            let blue = c[4..6].to_string();
            musthash = musthash.insert(b.to_string() + "-hex-b", blue.as_ref());
            let blue = i32::from_str_radix(c[4..6].as_ref(), 16).unwrap();
            musthash = musthash.insert(b.to_string() + "-rgb-b", blue);
            musthash = musthash.insert(b.to_string() + "-dec-b", blue as f64 / 255.0);

            musthash = musthash.insert(
                b.to_string() + "-hex-bgr",
                format!("{}{}{}", blue, green, red),
            );
        }

        if let Some(x) = Path::new(file.clone().as_str()).parent() {
            fs::create_dir_all(x)?
        };

        let f = File::create(file.clone())?;
        let mut out = BufWriter::new(f);
        musthash.render(&mustache, &mut out)
            .map_err(|_| io::Error::new(io::ErrorKind::InvalidData, "Invalid template format."))?;

        if Command::new("sh")
            .arg("-c")
            .arg(options
                .as_table()
                .unwrap_or(&Map::new())
                .get("post")
                .unwrap_or(&Value::String("".to_string()))
                .as_str()
                .unwrap_or(""))
            .spawn()
            .is_ok()
        {};
    };

    if Command::new("sh")
        .arg("-c")
        .arg(config.get("post")
        .unwrap_or(&Value::String("".to_string()))
        .as_str().unwrap_or(""))
        .spawn()
        .is_ok()
    {};

    Ok(())
}
